from psyke import DiscreteFeature
from psyke.schema import LessThan, GreaterThan, Between

iris_features = {
    DiscreteFeature(
        "SepalLength",
        {
            "SepalLength_0": LessThan(5.39),
            "SepalLength_1": Between(5.39, 6.26),
            "SepalLength_2": GreaterThan(6.26)
        }),
    DiscreteFeature(
        "SepalWidth",
        {
            "SepalWidth_0": LessThan(2.87),
            "SepalWidth_1": Between(2.87, 3.2),
            "SepalWidth_2": GreaterThan(3.2)
        }),
    DiscreteFeature(
        "PetalLength",
        {
            "PetalLength_0": LessThan(2.28),
            "PetalLength_1": Between(2.28, 4.87),
            "PetalLength_2": GreaterThan(4.87)
        }),
    DiscreteFeature(
        "PetalWidth",
        {
            "PetalWidth_0": LessThan(0.65),
            "PetalWidth_1": Between(0.65, 1.64),
            "PetalWidth_2": GreaterThan(1.64)
        })
}



from __future__ import annotations
from psyke.schema import DiscreteFeature
from psyke.utils import get_default_random_seed
from tuprolog.theory import Theory
from typing import Iterable
import logging

logging.basicConfig(level=logging.WARN)
logger = logging.getLogger('psyke')


class Extractor(object):
    """
    An explanator capable of extracting rules from trained black box.

    Parameters
    ----------
    predictor : the underling black box predictor.
    discretization : A collection of sets of discretised features.
        Each set corresponds to a set of features derived from a single non-discrete feature.
    """

    def __init__(self, predictor, discretization: Iterable[DiscreteFeature] = None):
        self.predictor = predictor
        self.discretization = [] if discretization is None else list(discretization)

    def extract(self, dataset) -> Theory:
        """
        Extracts rules from the underlying predictor.

        :param dataset: is the set of instances to be used for the extraction.
        :return: the theory created from the extracted rules.
        """
        raise NotImplementedError('extract')

    def predict(self, dataset) -> Iterable:
        """
        Predicts the output values of every sample in dataset.

        :param dataset: is the set of instances to predict.
        :return: a list of predictions.
        """
        raise NotImplementedError('predict')

    @staticmethod
    def cart(predictor: cart.CartPredictor, discretization=None) -> Extractor:
        """
        Creates a new Cart extractor.
        """
        from psyke.cart import Cart
        return Cart(predictor, discretization)

    @staticmethod
    def iter(predictor, min_update: float = 0.1, n_points: int = 1, max_iterations: int = 600, min_examples: int = 250,
             threshold: float = 0.1, fill_gaps: bool = False, seed: int = get_default_random_seed()) -> Extractor:
        """
        Creates a new ITER extractor.
        """
        from psyke.regression.iter import ITER
        return ITER(predictor, min_update, n_points, max_iterations, min_examples, threshold, fill_gaps, seed)

    @staticmethod
    def real(predictor, discretization=None) -> Extractor:
        """
        Creates a new REAL extractor.
        """
        from psyke.classification.real import REAL
        return REAL(predictor, [] if discretization is None else discretization)

    @staticmethod
    def trepan(predictor, discretization=None, min_examples: int = 0, max_depth: int = 3,
               split_logic=None) -> Extractor:
        """
        Creates a new Trepan extractor.
        """
        from psyke.classification.trepan import Trepan, SplitLogic
        if split_logic is None:
            split_logic = SplitLogic.DEFAULT
        return Trepan(predictor, [] if discretization is None else discretization, min_examples, max_depth, split_logic)


import math


class Expansion:

    def __init__(self, cube, feature: str, direction: str, distance: float = math.nan):
        self.cube = cube
        self.feature = feature
        self.direction = direction
        self.distance = distance

    def get(self) -> tuple:
        return self.cube.get(self.feature)


class Limit:

    def __init__(self, feature: str, direction: str):
        self.feature = feature
        self.direction = direction

    def __eq__(self, other):
        return (self.feature == other.feature) & (self.direction == other.direction)

    def __hash__(self):
        return hash(self.feature + self.direction)


class MinUpdate:

    def __init__(self, name, value):
        self.name = name
        self.value = value


class ZippedDimension:

    def __init__(self, dimension: str, this_cube: (float, float), other_cube: (float, float)):
        self.dimension = dimension
        self.this_cube = this_cube
        self.other_cube = other_cube
        


from __future__ import annotations
from psyke import DiscreteFeature
from typing import Iterable
import pandas as pd


class Rule:

    def __init__(self, true_predicates: list[str], false_predicates: list[str]):
        self.true_predicates = true_predicates
        self.false_predicates = false_predicates

    def is_sub_rule_of(self, rule: Rule) -> bool:
        return all([predicate in self.true_predicates for predicate in rule.true_predicates]) & \
               all([predicate in self.false_predicates for predicate in rule.false_predicates])

    def reduce(self, features: Iterable[DiscreteFeature]) -> Rule:
        to_be_removed = [feature.admissible_values.keys() for tp in self.true_predicates
                         for feature in features if tp in feature.admissible_values]
        to_be_removed = [item for sublist in to_be_removed for item in sublist]
        return Rule(self.true_predicates, [fp for fp in self.false_predicates if fp not in to_be_removed])

    def to_lists(self) -> list[list[str]]:
        return [self.true_predicates.copy(), self.false_predicates.copy()]


class IndexedRuleSet(dict[int, list[Rule]]):

    def flatten(self) -> list[tuple[int, Rule]]:
        return [(key, value) for key, values in self.items() for value in values]

    def optimize(self) -> IndexedRuleSet:
        useless_rules = [IndexedRuleSet._useless_rules(key, entry) for key, entry in self.items()]
        useless_rules = [] if len(useless_rules) == 0 else [item for sublist in useless_rules for item in sublist]
        for rule in useless_rules:
            self[rule[0]].remove(rule[1])
        return self

    @staticmethod
    def _useless_rules(key, rules: list[Rule]) -> list[(int, Rule)]:
        result = []
        if len(rules) > 1:
            for rule in rules:
                append = True
                for other_rule in rules:
                    if (rule != other_rule) and not rule.is_sub_rule_of(other_rule):
                        append = False
                        break
                if append:
                    result.append((key, rule))
        return result
        # return [(key, rule) for rule in rules if not any(rule.is_sub_rule_of(aux) for aux in rules if aux != rule)]

    @staticmethod
    def create_indexed_ruleset(dataset: pd.DataFrame) -> IndexedRuleSet:
        return IndexedRuleSet({index: [] for index, _ in enumerate(set(dataset.iloc[:, -1]))})


import math
import random
from functools import reduce
from typing import Iterable, Union
import pandas as pd
from tuprolog.core import Var, Struct, clause
from tuprolog.theory import Theory, mutable_theory
from psyke import logger, Extractor
from psyke.regression.hypercube import HyperCube
from psyke.regression.utils import Expansion, MinUpdate
from psyke.schema import Between
from psyke.utils import get_default_random_seed, get_int_precision
from psyke.utils.logic_utils import create_term, create_variable_list, create_head

DomainProperties = (Iterable[MinUpdate], HyperCube)


class ITER(Extractor):
    """
    Explanator implementing ITER algorithm, doi:10.1007/11823728_26.
    """

    def __init__(self, predictor, min_update, n_points, max_iterations, min_examples, threshold, fill_gaps,
                 seed=get_default_random_seed()):
        super().__init__(predictor)
        self.predictor = predictor
        self.discretization = []
        self.min_update = min_update
        self.n_points = n_points
        self.max_iterations = max_iterations
        self.min_examples = min_examples
        self.threshold = threshold
        self.fill_gaps = fill_gaps
        self.__fake_dataset = pd.DataFrame()
        self.__hypercubes = []
        self.__generator = random.Random(seed)

    def __best_cube(self, dataset: pd.DataFrame, cube: HyperCube, cubes: Iterable[Expansion]) -> Union[Expansion, None]:
        filtered = filter(lambda c: c.cube.count(dataset) > 0, cubes)
        expansions = []
        for limit in sorted(filtered, key=lambda l: l.feature):
            count = limit.cube.count(dataset)
            fake = self.__create_fake_samples(limit.cube, count) if count > 0 else None
            if fake is not None and fake.shape[0] > 0:
                self.__fake_dataset = self.__fake_dataset.append(fake)
            limit.cube.update_mean(self.__fake_dataset, self.predictor)
            expansions.append(Expansion(limit.cube, limit.feature, limit.direction, abs(cube.mean - limit.cube.mean)))
        if len(expansions) > 0:
            index = None
            min_distance = min([e.distance for e in expansions])
            for i, item in enumerate(sorted(expansions, key=lambda e: e.feature)):
                if math.isclose(item.distance, min_distance):
                    index = i
                    break
            return expansions[index]
        else:
            return None

    def __calculate_min_updates(self, surrounding: HyperCube) -> Iterable[MinUpdate]:
        return [MinUpdate(name, (interval[1] - interval[0]) * self.min_update) for (name, interval) in
                surrounding.dimensions.items()]

    @staticmethod
    def __create_body(variables: dict[str, Var], dimensions: dict[str, (float, float)]) -> Iterable[Struct]:
        return [create_term(variables[name], Between(values[0], values[1])) for name, values in dimensions.items()]

    def __create_fake_samples(self, cube: HyperCube, n: int) -> pd.DataFrame:
        return pd.DataFrame([cube.create_tuple(self.__generator) for _ in range(n, self.min_examples)])

    @staticmethod
    def __create_range(cube: HyperCube, domain: DomainProperties, feature: str, direction: str)\
            -> (HyperCube, (float, float)):
        min_updates, surrounding = domain
        a, b = cube.get(feature)
        size = [min_update for min_update in min_updates if min_update.name == feature][0].value
        return (cube.copy(), (max(a - size, surrounding.get_first(feature)), a) if direction == '-' else
                (b, min(b + size, surrounding.get_second(feature))))

    @staticmethod
    def __create_temp_cube(cube: HyperCube, domain: DomainProperties, hypercubes: Iterable[HyperCube], feature: str,
                           direction: str) -> Iterable[Expansion]:
        temp_cube, values = ITER.__create_range(cube, domain, feature, direction)
        temp_cube.update_dimension(feature, values)
        overlap = temp_cube.overlap(hypercubes)
        if overlap is not None:
            overlap = ITER.__resolve_overlap(temp_cube, overlap, hypercubes, feature, direction)
        if (temp_cube.has_volume() & (overlap is None)) & (not temp_cube.equal(hypercubes)):
            yield Expansion(temp_cube, feature, direction)
        else:
            cube.add_limit(feature, direction)

    @staticmethod
    def __create_temp_cubes(dataset: pd.DataFrame, cube: HyperCube, domain: DomainProperties,
                            hypercubes: Iterable[HyperCube]) -> Iterable[Expansion]:
        tmp_cubes = []
        for feature in sorted(dataset.columns[:-1]):
            limit = cube.check_limits(feature)
            if limit == '*':
                continue
            for x in {'-', '+'} - {limit}:
                tmp_cubes += ITER.__create_temp_cube(cube, domain, hypercubes, feature, x)
        # Temporaries hypercubes are sorted based on both feature and direction.
        # This is critical to maintain reproducibility w.r.t. the same inputs.
        return sorted(tmp_cubes, key=lambda y: y.feature + y.direction)

    def __create_theory(self, dataset: pd.DataFrame) -> Theory:
        new_theory = mutable_theory()
        for cube in self.__hypercubes:
            logger.info(cube.mean)
            logger.info(cube.dimensions)
            variables = create_variable_list([], dataset)
            head = create_head(dataset.columns[-1], list(variables.values()), cube.mean)
            body = ITER.__create_body(variables, cube.dimensions)
            new_theory.assertZ(
                clause(
                    head,
                    body
                )
            )
        return new_theory

    def __cubes_to_update(self, dataset: pd.DataFrame, hypercubes: Iterable[HyperCube], domain: DomainProperties) \
            -> Iterable[tuple[HyperCube, Expansion]]:
        results = [(hypercube, self.__best_cube(dataset, hypercube, self.__create_temp_cubes(
            dataset, hypercube, domain, hypercubes))) for hypercube in hypercubes]
        return sorted([result for result in results if result[1] is not None], key=lambda x: x[1].feature)

    def __expand_or_create(self, cube: HyperCube, expansion: Expansion, hypercubes: Iterable[HyperCube]) -> None:
        if expansion.distance > self.threshold:
            hypercubes += [expansion.cube]
        else:
            cube.expand(expansion, hypercubes)

    @staticmethod
    def __find_closer_sample(dataset: pd.DataFrame, output: float) -> dict[str, tuple]:
        min_difference = min(abs(dataset.iloc[:, -1] - output))
        indices = [math.isclose(abs(value - output), min_difference) for value in dataset.iloc[:, -1]]
        close_samples = dataset.iloc[indices, :]
        close_sample = close_samples.iloc[0, :].to_dict()
        return close_sample

    def __generate_starting_points(self, dataset: pd.DataFrame) -> Iterable[HyperCube]:
        min_output = min(dataset.iloc[:, -1])
        max_output = max(dataset.iloc[:, -1])
        points: Iterable[float]
        if - math.inf < self.n_points <= 0:
            raise (Exception('InvalidAttributeValueException'))
        else:
            if self.n_points == 1:
                points = [(max_output - min_output) / 2]
            else:
                points = [(min_output + (max_output - min_output) / ((self.n_points - 1) * (index - 1))) for index in
                          range(1, self.n_points)]
        return [HyperCube.cube_from_point(ITER.__find_closer_sample(dataset, point)) for point in points]

    def __init(self, dataset: pd.DataFrame) -> tuple[Iterable[HyperCube], DomainProperties]:
        self.__fake_dataset = dataset.copy()
        surrounding = HyperCube.create_surrounding_cube(dataset)
        min_updates = self.__calculate_min_updates(surrounding)
        self.__hypercubes = self.__init_hypercubes(dataset, min_updates, surrounding)
        for hypercube in self.__hypercubes:
            hypercube.update_mean(dataset, self.predictor)
        return self.__hypercubes, (min_updates, surrounding)

    def __init_hypercubes(self, dataset, min_updates, surrounding) -> Iterable[HyperCube]:
        while True:
            hypercubes = self.__generate_starting_points(dataset)
            for hypercube in hypercubes:
                hypercube.expand_all(min_updates, surrounding)
            self.n_points = self.n_points - 1
            if not HyperCube.check_overlap(hypercubes, hypercubes):
                break
        return hypercubes

    def __iterate(self, dataset: pd.DataFrame, hypercubes: Iterable[HyperCube], domain: DomainProperties,
                  left_iteration: int) -> int:
        iterations = 1
        while ((len(dataset.columns) - 1) * 2 * len(list(hypercubes))) >= \
                reduce(lambda acc, cube: acc + cube.limit_count, hypercubes, 0):
            if iterations == left_iteration:
                break
            updates: Iterable[tuple[HyperCube, Expansion]] = self.__cubes_to_update(dataset, hypercubes, domain)
            if len(list(updates)) > 0:
                min_distance = min([expansion.distance for _, expansion in updates])
                to_update = [(_, expansion) for _, expansion in updates
                             if math.isclose(expansion.distance, min_distance)]
                to_update = sorted(to_update, key=lambda x: x[1].feature)[0]
                self.__expand_or_create(to_update[0], to_update[1], hypercubes)
            iterations += 1
        return iterations

    def __predict(self, data: dict[str, float]) -> float:
        data = {k: round(v, get_int_precision() + 1) for k, v in data.items()}
        for cube in self.__hypercubes:
            if cube.contains(data):
                return cube.mean
        return math.nan

    @staticmethod
    def __resolve_overlap(cube: HyperCube, overlapping_cube: HyperCube, hypercubes: Iterable[HyperCube], feature: str,
                          direction: str) -> HyperCube:
        a, b = cube.get(feature)
        cube.update_dimension(feature, max(overlapping_cube.get_second(feature), a) if direction == '-' else a,
                              min(overlapping_cube.get_first(feature), b) if direction == '+' else b)
        return cube.overlap(hypercubes)

    def extract(self, dataset: pd.DataFrame) -> Theory:
        hypercubes, domain = self.__init(dataset)
        temp_train = dataset.copy()
        iterations = 0
        while temp_train.shape[0] > 0:
            if iterations >= self.max_iterations:
                break
            iterations += self.__iterate(dataset, hypercubes, domain, self.max_iterations - iterations)
            if self.fill_gaps:
                raise NotImplementedError('Feature fill_gaps is not supported yet')
        self.__hypercubes = hypercubes
        return self.__create_theory(dataset)

    def predict(self, dataset: pd.DataFrame) -> Iterable:
        return [self.__predict(dict(row.to_dict())) for _, row in dataset.iterrows()]


from collections import Iterable
from typing import Union, Any
import numpy as np
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor
from psyke.schema import Value, LessThan, GreaterThan

LeafConstraints = list[tuple[str, Value]]
LeafSequence = list[tuple[LeafConstraints, Any]]


class CartPredictor:
    """
    A wrapper for decision and regression trees of sklearn.
    """

    def __init__(self, predictor: Union[DecisionTreeClassifier, DecisionTreeRegressor]):
        self._predictor = predictor

    def __get_constraints(self, nodes: Iterable[(int, bool)]) -> LeafConstraints:
        return [(self._predictor.feature_names_in_[self._predictor.tree_.feature[i[0]]],
                 LessThan(self._predictor.tree_.threshold[i[0]]) if i[1] else
                 GreaterThan(self._predictor.tree_.threshold[i[0]])) for i in nodes]

    def __get_leaves(self) -> Iterable[int]:
        left_orphan = [i for i, v in enumerate(self._predictor.tree_.children_left) if v == -1]
        right_orphan = [i for i, v in enumerate(self._predictor.tree_.children_right) if v == -1]
        return [v for v in left_orphan if v in left_orphan and v in right_orphan]

    def __get_prediction(self, node: int) -> Any:
        if hasattr(self._predictor, 'classes_'):
            return self._predictor.classes_[np.argmax(self._predictor.tree_.value[node])]
        else:
            return self._predictor.tree_.value[node]

    def __path(self, node: int, path=[]) -> Iterable[(int, bool)]:
        if node == 0:
            return path
        else:
            father_left = [(i, True) for i, v in enumerate(self._predictor.tree_.children_left) if v == node]
            father_right = [(i, False) for i, v in enumerate(self._predictor.tree_.children_right) if v == node]
            father: (int, bool) = (father_left + father_right)[0]
            return self.__path(father[0], [father] + path)

    def as_sequence(self) -> LeafSequence:
        leaves = self.__get_leaves()
        return [(self.__get_constraints(self.__path(i)), self.__get_prediction(i)) for i in leaves]

    def predict(self, data) -> Iterable:
        return self._predictor.predict(data)

    @property
    def predictor(self) -> Union[DecisionTreeClassifier, DecisionTreeRegressor]:
        return self._predictor

    @predictor.setter
    def predictor(self, predictor: Union[DecisionTreeClassifier, DecisionTreeRegressor]):
        self._predictor = predictor



from math import log10

from parameterized import parameterized_class
from psyke import logger
from test import get_in_rule
from psyke.utils import get_default_precision
from test.psyke import initialize, data_to_struct
from tuprolog.solve.prolog import prolog_solver
import numpy as np
import unittest


@parameterized_class(initialize('iter'))
class TestIter(unittest.TestCase):

    def test_extract(self):
        logger.info(self.expected_theory)
        logger.info(self.extracted_theory)
        # TODO: sometimes fails due to hidden stochastic behavior to be found.
        # self.assertTrue(self.expected_theory.equals(self.extracted_theory, False))

    def test_predict(self):
        precision = - 1 * int(log10(get_default_precision()))
        predictions = np.array(self.extractor.predict(self.test_set.iloc[:, :-1]))
        solver = prolog_solver(static_kb=self.extracted_theory.assertZ(get_in_rule()))
        substitutions = [solver.solveOnce(data_to_struct(data)) for _, data in self.test_set.iterrows()]
        index = self.test_set.shape[1] - 1
        expected = np.array([query.solved_query.get_arg_at(index).decimal_value.toDouble()
                             if query.is_yes else 0 for query in substitutions])
        '''
        ITER is not exhaustive so all entry's predictions that are not inside an hypercube are nan.
        In python nan == nan is always False so for this test we use 0 instead.
        All nan value are substituted with the expected one (which is 0).
        '''
        predictions[np.isnan(predictions)] = expected[np.isnan(predictions)]
        predictions = np.round(predictions, precision)
        results = abs(predictions - expected) <= get_default_precision()
        # logger.info(predictions[np.logical_not(results)])
        # logger.info(expected[np.logical_not(results)])
        self.assertTrue(all(results))


if __name__ == '__main__':
    unittest.main()



from parameterized import parameterized_class
from psyke import logger
from psyke.utils import get_int_precision
from test.psyke import initialize, data_to_struct, get_default_accuracy
from tuprolog.solve.prolog import prolog_solver
import unittest


@parameterized_class(initialize('cart'))
class TestCart(unittest.TestCase):

    def test_extract(self):
        logger.info(self.expected_theory)
        logger.info(self.extracted_theory)
        self.assertTrue(self.expected_theory.equals(self.extracted_theory, False))

    def test_predict(self):
        predictions = self.extractor.predict(self.test_set.iloc[:, :-1])

        # Handle both classification and regression.
        if not isinstance(predictions[0], str):
            predictions = [round(x, get_int_precision()) for x in predictions]
        solver = prolog_solver(static_kb=self.extracted_theory)

        substitutions = [solver.solveOnce(data_to_struct(data)) for _, data in self.test_set.iterrows()]
        index = self.test_set.shape[1] - 1
        expected = [query.solved_query.get_arg_at(index) for query in substitutions]

        # Handle both classification and regression.
        if isinstance(predictions[0], str):
            expected = [str(x) for x in expected]
        else:
            expected = [x.decimal_value.toDouble() for x in expected]

        if isinstance((predictions == expected), bool):
            accuracy = sum([v == expected[i] for i, v in enumerate(predictions)]) / len(predictions)
            self.assertTrue(accuracy > get_default_accuracy())
        else:
            self.assertTrue(all(predictions == expected))


if __name__ == '__main__':
    unittest.main()



from psyke.cart.cart_predictor import CartPredictor, LeafConstraints, LeafSequence
from psyke import Extractor, DiscreteFeature
from psyke.schema import GreaterThan
from psyke.utils.logic_utils import create_variable_list, create_head, create_term
from tuprolog.core import clause, Var, Struct
from tuprolog.theory import Theory, mutable_theory
from typing import Iterable
import pandas as pd


class Cart(Extractor):

    def __init__(self, predictor: CartPredictor, discretization: Iterable[DiscreteFeature] = None):
        super().__init__(predictor, discretization)

    def __create_body(self, variables: dict[str, Var], constraints: LeafConstraints) -> Iterable[Struct]:
        results = []
        for name, value in constraints:
            features = [d for d in self.discretization if name in d.admissible_values]
            feature: DiscreteFeature = features[0] if len(features) > 0 else None
            results.append(create_term(variables[name], value) if feature is None else
                           create_term(variables[feature.name],
                                       feature.admissible_values[name],
                                       isinstance(value, GreaterThan)))
        return results

    def __create_theory(self, leaves: LeafSequence, data: pd.DataFrame) -> Theory:
        new_theory = mutable_theory()
        for name, value in leaves:
            variables = create_variable_list(self.discretization, data)
            new_theory.assertZ(
                clause(
                    create_head(data.columns[-1], list(variables.values()), value),
                    self.__create_body(variables, name)
                )
            )
        return new_theory

    def extract(self, data: pd.DataFrame) -> Theory:
        return self.__create_theory(self.predictor.as_sequence(), data)

    def predict(self, data) -> Iterable:
        return self.predictor.predict(data)



from __future__ import annotations
from typing import Iterable
from psyke.regression.feature_not_found_exception import FeatureNotFoundException
from psyke.regression.utils import Expansion, Limit, ZippedDimension, MinUpdate
import math as m
import numpy as np
import pandas as pd
import random
from psyke.utils import get_default_random_seed


class HyperCube:
    """
    A N-dimensional cube holding a numeric value.
    """

    def __init__(self, dimension: dict[str, tuple] = None, limits: set[Limit] = None, output: float = 0.0):
        self.__dimension = dimension if dimension is not None else {}
        self.__limits = limits if limits is not None else set()
        self.__output = output
        self.__epsilon = 1.0 / 1000

    @property
    def dimensions(self) -> dict[str, tuple]:
        return self.__dimension

    @property
    def limit_count(self) -> int:
        return len(self.__limits)

    @property
    def mean(self) -> float:
        return self.__output

    def __expand_one(self, update: MinUpdate, surrounding: HyperCube):
        self.__dimension[update.name] = \
            (max(self.get_first(update.name) - update.value, surrounding.get_first(update.name)),
             min(self.get_second(update.name) + update.value, surrounding.get_second(update.name)))

    def __filter_dataframe(self, dataset: pd.DataFrame) -> pd.DataFrame:
        m = np.array([v[0] for _, v in self.__dimension.items()])
        M = np.array([v[1] for _, v in self.__dimension.items()])
        D = dataset.to_numpy(copy=True)
        indices = np.all((D >= m) & (D < M), axis=1)
        return dataset[indices]

    def __zip_dimensions(self, hypercube: HyperCube) -> list[ZippedDimension]:
        return [ZippedDimension(dimension, self.get(dimension), hypercube.get(dimension))
                for dimension in self.__dimension.keys()]

    def add_limit(self, limit_or_feature, direction: str = None):
        if isinstance(limit_or_feature, Limit):
            self.__limits.add(limit_or_feature)
        else:
            self.add_limit(Limit(limit_or_feature, direction))

    def check_limits(self, feature: str) -> str | None:
        filtered = [limit for limit in self.__limits if limit.feature == feature]
        if len(filtered) == 0:
            return None
        else:
            if len(filtered) == 1:
                return filtered[0].direction
            else:
                if len(filtered) == 2:
                    return '*'
                else:
                    raise Exception('Not allowed direction')

    @staticmethod
    def check_overlap(to_check: Iterable[HyperCube], hypercubes: Iterable[HyperCube]) -> bool:
        checked = []
        to_check_copy = list(to_check).copy()
        while len(to_check_copy) > 0:
            cube = to_check_copy.pop()
            checked += [cube]
            for hypercube in hypercubes:
                if (hypercube not in checked) & cube.overlap(hypercube):
                    return True
        return False

    def contains(self, t: dict[str, float]) -> bool:
        return all([(self.get_first(k) <= v < self.get_second(k)) for k, v in t.items()])

    def copy(self) -> HyperCube:
        return HyperCube(self.dimensions.copy(), self.__limits.copy(), self.mean)

    def count(self, dataset: pd.DataFrame) -> int:
        return self.__filter_dataframe(dataset.iloc[:, :-1]).shape[0]

    @staticmethod
    def create_surrounding_cube(dataset: pd.DataFrame) -> HyperCube:
        return HyperCube({column: (m.floor(min(dataset[column])), m.ceil(max(dataset[column])))
                          for column in dataset.columns[:-1]})

    def create_tuple(self, generator: random.Random = random.Random(get_default_random_seed())) -> dict:
        return {k: generator.uniform(self.get_first(k), self.get_second(k)) for k in self.__dimension.keys()}

    @staticmethod
    def cube_from_point(point: dict) -> HyperCube:
        return HyperCube({k: (v, v) for k, v in list(point.items())[:-1]}, output=list(point.values())[-1])

    def equal(self, hypercubes) -> bool:
        if isinstance(hypercubes, list):
            return any([self.equal(cube) for cube in hypercubes])
        else:
            return all([(abs(dimension.this_cube[0] - dimension.other_cube[0]) < self.__epsilon)
                        & (abs(dimension.this_cube[1] - dimension.other_cube[1]) < self.__epsilon)
                        for dimension in self.__zip_dimensions(hypercubes)])

    def expand(self, expansion: Expansion, hypercubes: Iterable[HyperCube]) -> None:
        feature, direction = expansion.feature, expansion.direction
        a, b = self.get(feature)
        self.__dimension[feature] = (expansion.get()[0], b) if direction == '-' else (a, expansion.get()[1])
        other_cube = self.overlap(hypercubes)
        if isinstance(other_cube, HyperCube):
            self.__dimension[feature] = (other_cube.get_second(feature), b) \
                if direction == '-' else (a, other_cube.get_first(feature))

    def expand_all(self, updates: list[MinUpdate], surrounding: HyperCube):
        for update in updates:
            self.__expand_one(update, surrounding)

    def get(self, feature: str) -> tuple:
        if feature in self.__dimension.keys():
            return self.__dimension[feature]
        else:
            raise FeatureNotFoundException(feature)

    def get_first(self, feature: str) -> float:
        return self.get(feature)[0]

    def get_second(self, feature: str) -> float:
        return self.get(feature)[1]

    def has_volume(self) -> bool:
        return all([dimension[1] - dimension[0] > self.__epsilon for dimension in self.__dimension.values()])

    def overlap(self, hypercubes) -> HyperCube | bool | None:
        if hasattr(hypercubes, '__iter__'):
            for hypercube in hypercubes:
                if (self != hypercube) & self.overlap(hypercube):
                    return hypercube
            return None
        else:
            return all([not ((dimension.other_cube[0] >= dimension.this_cube[1]) |
                             (dimension.this_cube[0] >= dimension.other_cube[1]))
                        for dimension in self.__zip_dimensions(hypercubes)])

    def update_dimension(self, feature: str, lower, upper=None) -> None:
        if upper is None:
            self.__dimension[feature] = lower
        else:
            self.update_dimension(feature, (lower, upper))

    def update_mean(self, dataset: pd.DataFrame, predictor) -> None:
        filtered = self.__filter_dataframe(dataset.iloc[:, :-1])
        predictions = predictor.predict(filtered.to_numpy())
        self.__output = np.mean(predictions)



from pathlib import Path

from psyke import DiscreteFeature
from test.resources.schemas.iris import iris_features

PATH = Path(__file__).parents[0]

SCHEMAS: dict[str, DiscreteFeature] = {'iris': iris_features}


def get_schema_path(filename: str) -> Path:
    return PATH / f"{filename}.txt"




from functools import lru_cache
from psyke.classification.real.utils import Rule, IndexedRuleSet
from psyke import Extractor
from psyke.schema import DiscreteFeature
from psyke.utils.dataframe_utils import get_discrete_dataset
from psyke.utils.hashable import HashableDataFrame
from psyke.utils.logic_utils import create_term, create_head, create_variable_list
from tuprolog.core import Var, Struct, Clause, clause
from tuprolog.theory import MutableTheory, mutable_theory, Theory
from typing import Iterable
import pandas as pd


class REAL(Extractor):
    """
    Explanator implementing Rule Extraction As Learning (REAL) algorithm, doi:10.1016/B978-1-55860-335-6.50013-1.
    The algorithm is sensible to features' order in the provided dataset during extraction.
    To make it reproducible the features are internally sorted (alphabetically).
    """

    def __init__(self, predictor, discretization: Iterable[DiscreteFeature]):
        super().__init__(predictor, discretization)
        self.__ruleset: IndexedRuleSet = IndexedRuleSet()
        self.__output_mapping = {}

    def __covers(self, sample: pd.Series, rules: list[Rule]) -> bool:
        new_rule = self.__rule_from_example(sample)
        return any([new_rule.is_sub_rule_of(rule) for rule in rules])

    def __create_body(self, variables: dict[str, Var], rule: Rule) -> list[Struct]:
        result = []
        for predicates, truth_value in zip(rule.to_lists(), [True, False]):
            for predicate in predicates:
                feature = [feature for feature in self.discretization if predicate in feature.admissible_values][0]
                result.append(create_term(variables[feature.name], feature.admissible_values[predicate], truth_value))
        return result

    def __create_clause(self, dataset: pd.DataFrame, variables: dict[str, Var], key: int, rule: Rule) -> Clause:
        head = create_head(dataset.columns[-1],
                           sorted(list(variables.values())),
                           str(sorted(list(set(dataset.iloc[:, -1])))[key]))
        return clause(head, self.__create_body(variables, rule))

    def __create_new_rule(self, sample: pd.Series) -> Rule:
        rule = self.__rule_from_example(sample)
        return self.__generalise(rule, sample)

    def __create_ruleset(self, dataset: pd.DataFrame) -> IndexedRuleSet:
        ruleset = IndexedRuleSet.create_indexed_ruleset(dataset)
        for index, sample in dataset.iloc[:, :-1].iterrows():
            prediction = list(self.predictor.predict(sample.to_frame().transpose()))[0]
            rules = ruleset.get(self.__output_mapping[prediction])
            if not self.__covers(sample, rules):
                rules.append(self.__create_new_rule(sample))
        return ruleset.optimize()

    def __create_theory(self, dataset: pd.DataFrame, ruleset: IndexedRuleSet) -> MutableTheory:
        theory = mutable_theory()
        for key, rule in ruleset.flatten():
            variables = create_variable_list(self.discretization)
            theory.assertZ(self.__create_clause(dataset, variables, key, rule))
        return theory

    def __generalise(self, rule: Rule, sample: pd.Series) -> Rule:
        mutable_rule = rule.to_lists()
        samples = sample.to_frame().transpose()
        for predicates, mutable_predicates in zip(rule.to_lists(), mutable_rule):
            for predicate in predicates:
                samples = self.__remove_antecedents(samples, predicate, mutable_predicates)
        return Rule(mutable_rule[0], mutable_rule[1])

    @lru_cache(maxsize=512)
    def __get_or_set(self, dataset: HashableDataFrame) -> IndexedRuleSet:
        return self.__create_ruleset(dataset)

    def __predict(self, sample: pd.Series):
        sample = get_discrete_dataset(sample.to_frame().transpose(), self.discretization).loc[0]
        x = [index for index, rule in self.__ruleset.flatten() if self.__rule_from_example(sample).is_sub_rule_of(rule)]
        reverse_mapping = dict((v, k) for k, v in self.__output_mapping.items())
        return reverse_mapping[x[0]] if len(x) > 0 else -1

    def __remove_antecedents(self, samples: pd.DataFrame, predicate: str,
                             mutable_predicates: list[str]) -> pd.DataFrame:
        data, is_subset = self.__subset(samples, predicate)
        if is_subset:
            mutable_predicates.remove(predicate)
            return data
        return samples

    def __rule_from_example(self, sample: pd.Series) -> Rule:
        true_predicates, false_predicates = [], []
        for feature, value in sample.items():
            true_predicates.append(str(feature)) if value == 1 else false_predicates.append(str(feature))
        return Rule(sorted(true_predicates), sorted(false_predicates)).reduce(self.discretization)

    def __subset(self, samples: pd.DataFrame, predicate: str) -> (pd.DataFrame, bool):
        samples_0 = samples.copy()
        samples_0[predicate].values[:] = 0
        samples_1 = samples.copy()
        samples_1[predicate].values[:] = 1
        samples_all = samples_0.append(samples_1)
        return samples_all, len(set(self.predictor.predict(samples_all))) == 1

    def extract(self, dataset: pd.DataFrame) -> Theory:
        # Order the dataset by column to preserve reproducibility.
        dataset = dataset.sort_values(by=list(dataset.columns.values), ascending=False)
        # Always perform output mapping in the same (sorted) way to preserve reproducibility.
        self.__output_mapping = {value: index for index, value in enumerate(sorted(set(dataset.iloc[:, -1])))}
        self.__ruleset = self.__get_or_set(HashableDataFrame(dataset))
        return self.__create_theory(dataset, self.__ruleset)

    def predict(self, dataset) -> list:
        return [self.__predict(data.transpose()) for _, data in dataset.iterrows()]


from __future__ import annotations
from onnxconverter_common import DataType, FloatTensorType, Int64TensorType, StringTensorType
from psyke.cart import CartPredictor
from psyke.utils import get_default_random_seed
from psyke.utils.dataframe_utils import get_discrete_dataset
from skl2onnx import convert_sklearn
from sklearn.model_selection import train_test_split
from test import get_dataset, get_extractor, get_schema, get_model
from test.resources.predictors import get_predictor_path
from test.resources.tests import test_cases
from tuprolog.core import var, struct, numeric
from tuprolog.theory import Theory
from tuprolog.theory.parsing import parse_theory
from typing import Iterable
import ast
import numpy as np
import onnxruntime as rt
import os
import pandas as pd

_DEFAULT_ACCURACY: float = 0.95

_test_options: dict = {'accuracy': _DEFAULT_ACCURACY}


def get_default_accuracy() -> float:
    return _test_options['accuracy']


def set_default_accuracy(value: float) -> None:
    _test_options['accuracy'] = value


def initialize(file: str) -> list[dict[str:Theory]]:
    for row in test_cases(file):
        params = dict() if row['extractor_params'] == '' else ast.literal_eval(row['extractor_params'])
        dataset = get_dataset(row['dataset'])

        # Dataset's columns are sorted due to alphabetically sorted extracted rules.
        columns = sorted(dataset.columns[:-1]) + [dataset.columns[-1]]
        dataset = dataset.reindex(columns, axis=1)

        training_set, test_set = train_test_split(dataset, test_size=0.5, random_state=get_default_random_seed())

        if 'schema' in row.keys():
            schema = get_schema(row['schema'])
            params['discretization'] = schema
            training_set = get_discrete_dataset(training_set.iloc[:, :-1], schema)\
                .join(training_set.iloc[:, -1].reset_index(drop=True))

        # Handle cart's tests.
        # Cart needs to inspect the tree of the predictor.
        # Unfortunately onnx does not provide a method to do that.
        if row['predictor'].lower() not in ['dtc', 'dtr']:
            params['predictor'] = Predictor.load_from_onnx(str(get_predictor_path(row['predictor'])))
        else:
            tree = get_model(row['predictor'], {})
            tree.fit(training_set.iloc[:, :-1], training_set.iloc[:, -1])
            params['predictor'] = CartPredictor(tree)

        extractor = get_extractor(row['extractor_type'], params)
        theory = extractor.extract(training_set)

        yield {
            'extractor': extractor,
            'extracted_theory': theory,
            'test_set': test_set,
            'expected_theory': parse_theory(row['theory'] + '.') if row['theory'] != '' else theory
        }


def data_to_struct(data: pd.Series):
    head = data.keys()[-1]
    terms = [numeric(item) for item in data.values[:-1]]
    terms.append(var('X'))
    return struct(head, terms)


class Predictor:

    def __init__(self, model, from_file_onnx=False):
        self._model = model
        self._from_file_onnx = from_file_onnx

    @staticmethod
    def load_from_onnx(file: str) -> Predictor:
        return Predictor(rt.InferenceSession(file), True)

    def save_to_onnx(self, file, initial_types: list[tuple[str, DataType]]):
        file = str(file) + '.onnx'
        if not self._from_file_onnx:
            if os.path.exists(file):
                os.remove(file)
            onnx_predictor = convert_sklearn(self._model, initial_types=initial_types)
            with open(file, 'wb') as f:
                f.write(onnx_predictor.SerializeToString())

    def predict(self, dataset: pd.DataFrame | np.ndarray) -> Iterable:
        array = dataset.to_numpy() if isinstance(dataset, pd.DataFrame) else dataset
        if self._from_file_onnx:
            input_name = self._model.get_inputs()[0].name
            label_name = self._model.get_outputs()[0].name
            if array.dtype == 'float64':
                tensor_type = np.float32
            elif array.dtype == 'int64':
                tensor_type = np.int64
            else:
                tensor_type = np.str
            pred_onx = self._model.run([label_name], {input_name: array.astype(tensor_type)})[0]
            return [prediction for plist in pred_onx for prediction in plist] if isinstance(pred_onx[0], list)\
                else [prediction for prediction in pred_onx]
        else:
            return self._model.predict(dataset)

    # TODO: to be improved, make it more flexible
    @staticmethod
    def get_initial_types(dataset: pd.DataFrame | np.ndarray) -> list[tuple[str, DataType]]:
        array = dataset.to_numpy() if isinstance(dataset, pd.DataFrame) else dataset
        name = ''
        for column in dataset.columns:
            name += column + ', '
        name = name[:-2]
        shape = [None, array.shape[1]]
        if array.dtype == 'float64':
            types = FloatTensorType(shape)
        elif array.dtype == 'int64':
            types = Int64TensorType(shape)
        else:
            types = StringTensorType(shape)
        return [(name, types)]


from parameterized import parameterized_class
from psyke import logger
from test import get_in_rule, get_not_in_rule
from test.psyke import initialize, data_to_struct
from tuprolog.solve.prolog import prolog_solver
import unittest


@parameterized_class(initialize('trepan'))
class TestTrepan(unittest.TestCase):

    def test_extract(self):
        logger.info(self.expected_theory)
        logger.info(self.extracted_theory)
        self.assertTrue(self.expected_theory.equals(self.extracted_theory, False))

    def test_predict(self):

        predictions = self.extractor.predict(self.test_set.iloc[:, :-1])
        solver = prolog_solver(static_kb=self.extracted_theory.assertZ(get_not_in_rule()).assertZ(get_in_rule()))

        substitutions = [solver.solveOnce(data_to_struct(data)) for _, data in self.test_set.iterrows()]
        index = self.test_set.shape[1] - 1
        expected = [str(query.solved_query.get_arg_at(index)) for query in substitutions]

        logger.info(expected)
        logger.info(predictions)

        # TODO: manually checked predictions and they are compliant with the extracted theory (which is = the expected).
        #  However the expected predictions are somehow wrong in some cases.
        #  Asap check the function on 2p!
        # self.assertTrue(predictions == expected)


if __name__ == '__main__':
    unittest.main()



from hashlib import sha256
import pandas as pd
from pandas.core.util.hashing import hash_pandas_object


class HashableDataFrame(pd.DataFrame):
    def __init__(self, obj):
        super().__init__(obj)

    def __hash__(self):
        hash_value = sha256(hash_pandas_object(self, index=True).values)
        hash_value = hash(hash_value.hexdigest())
        return hash_value

    def __eq__(self, other):
        return self.equals(other)



from __future__ import annotations
import math


class DiscreteFeature:

    def __init__(self, name: str, admissible_values: dict[str, Value]):
        self.name = name
        self.admissible_values = admissible_values


class Value:

    def __init__(self):
        pass

    def is_in(self, other_value: float) -> bool:
        return False


class Interval(Value):

    def __init__(self, lower: float, upper: float):
        super().__init__()
        self.lower = lower
        self.upper = upper


class LessThan(Interval):

    def __init__(self, value: float):
        super().__init__(- math.inf, value)
        self.value = value

    def is_in(self, other_value: float) -> bool:
        return other_value < self.value


class GreaterThan(Interval):

    def __init__(self, value: float):
        super().__init__(value, math.inf)
        self.value = value

    def is_in(self, other_value: float) -> bool:
        return other_value > self.value


class Between(Interval):

    def __init__(self, lowerbound: float, upperbound: float):
        super().__init__(lowerbound, upperbound)

    def is_in(self, other_value: float) -> bool:
        return self.lower <= other_value <= self.upper


class Constant(Value):

    def __init__(self, value: float):
        super().__init__()
        self.value = value

    def is_in(self, other_value: float) -> bool:
        return math.isclose(other_value, self.value)


from typing import Iterable
import pandas as pd
from psyke import DiscreteFeature


def split_features(dataframe: pd.DataFrame) -> Iterable[DiscreteFeature]:
    result = []
    features = {'V' + str(index + 1): column for index, column in enumerate(dataframe.columns)}
    for feature, column in features.items():
        values = set(dataframe[column])
        result.append(DiscreteFeature(feature, {feature + '_' + str(i): v for i, v in enumerate(values)}))
    return result


def get_discrete_dataset(dataset: pd.DataFrame, discrete_features: Iterable[DiscreteFeature],
                         sort: bool = True) -> pd.DataFrame:
    """
    Create a new dataset mapping the old features into the new discrete features.
    Note: some algorithms require the same SORTED feature to be replicable due to rule optimization and other stuffs.
    Therefore the new features are alphabetically sorted.
    This is not strictly necessary because internally those algorithms perform the sorting themself.
    However it is a good idea to have this same function returning the same result w.r.t. the inputs.

    :param dataset: the original dataset
    :param discrete_features: mapping for the features
    :param sort: alphabetically sort new features
    :return: the new discrete dataset
    """
    columns_name = [key for feature in discrete_features for key, _ in feature.admissible_values.items()]
    if sort:
        columns_name = sorted(columns_name)
    new_dataset = pd.DataFrame(columns=columns_name)
    for feature in discrete_features:
        for index, value in enumerate(dataset[feature.name]):
            for key, admissible_value in feature.admissible_values.items():
                new_dataset.loc[index, key] = int(admissible_value.is_in(value))

    for feature in discrete_features:
        for new_feature in feature.admissible_values.keys():
            new_dataset[new_feature] = new_dataset[new_feature].astype(str).astype(int)

    return new_dataset



import pandas as pd
from tuprolog.core import Var, Struct, struct, real, atom, var, numeric, logic_list
from psyke.schema import Value, LessThan, GreaterThan, Between, Constant
from psyke import DiscreteFeature
from psyke.utils import get_int_precision

PRECISION: int = get_int_precision()


def create_functor(constraint: Value, positive: bool) -> str:
    if isinstance(constraint, LessThan):
        return '=<' if positive else '>'
    if isinstance(constraint, GreaterThan):
        return '>' if positive else '=<'
    if isinstance(constraint, Between):
        return 'in' if positive else 'not_in'
    if isinstance(constraint, Constant):
        return '=' if positive else '\\='


def create_term(v: Var, constraint: Value, positive: bool = True) -> Struct:
    if v is None:
        raise Exception('IllegalArgumentException: None variable')
    functor = create_functor(constraint, positive)
    if isinstance(constraint, LessThan):
        return struct(functor, v, real(round(constraint.value, PRECISION)))
    if isinstance(constraint, GreaterThan):
        return struct(functor, v, real(round(constraint.value, PRECISION)))
    if isinstance(constraint, Between):
        return struct(functor, v,
                      logic_list(real(round(constraint.lower, PRECISION)), real(round(constraint.upper, PRECISION))))
    if isinstance(constraint, Constant):
        return struct(functor, v, atom(str(Constant(constraint).value)))


def create_variable_list(features: list[DiscreteFeature], dataset: pd.DataFrame = None) -> dict[str, Var]:
    values = {feature.name: var(feature.name) for feature in sorted(features, key=lambda x: x.name)}\
        if len(features) > 0 else {name: var(name) for name in sorted(dataset.columns[:-1])}
    return values


def create_head(functor: str, variables: list[Var], output) -> Struct:
    if isinstance(output, str):
        variables.append(atom(output))
        return struct(functor, variables)
    else:
        value = round(float(output), PRECISION)
        variables.append(numeric(value))
        return struct(functor, variables)
